import optuna


def sample_from_feasible_domain(trial: optuna.trial.BaseTrial, inputs, pre_name=""):
    """
        使用试验 trial 基于输入中的定义域 feasible_domain 部分进行参数采样和替换。
            遍历输入中的所有元素，找出符合 <feasible_domain> 格式要求的记录了参数定义域的元素，
            然后使用输入的试验实例 trial 结合参数定义域采样出对应的参数，最后用采样出来的参数替换掉原来的定义域。

        参数：
            inputs:             <list/dict> 当其中的元素满足 <feasible_domain> 格式要求，将进行采样与替换。
                                    <feasible_domain> 格式要求：
                                        1. 是一个 dictionary
                                        2. 包含 "p_type"  字段
                                            "p_type" 表示定义域类型，常见值包括："float" "int" "categorical" 等
                                        3. 根据不同的定义域类型，给出定义域的参数 
                                            比如，p_type="categorical" 时应该包含可选值列表 "choices"
                                            p_type="float"或者"int" 时应该包含最大、最小、间隔值、坐标轴类型 "high" "low" "step" "log" 等
                                        更多参见 https://optuna.readthedocs.io/zh_CN/latest/reference/generated/optuna.trial.Trial.html#optuna.trial.Trial
                                        中的 suggest_xxx() 函数。
            trial:              <optuna.trial.BaseTrial> 试验
            pre_name:           <string> 采样出的参数在试验 trial 中注册的名称的前缀
        实例：
            对于输入 inputs={
                            "thr":[
                                "interval_thr": {
                                    "p_type": "categorical",
                                    "choices": [
                                      1000,
                                      2000,
                                      5000,
                                      10000
                                    ]
                                },
                                "iou_thr": {
                                    "p_type": "float",
                                    "low": 0,
                                    "high": 1.0000001,
                                    "step": 0.05
                                },
                            ]
                        }
            可能返回的采样结果是 res={"thr":[{"interval_thr":1000}, {"iou_thr":0.6}], }。
            当 pre_name="my" 时，这些参数在 trial 中注册的名称分别是 "my:thr@0:interval_thr" 和 "my:thr@1:iou_thr"。
            这些名称的含义详见 get_value_by_name()。
    """
    if isinstance(inputs, (dict,)) and "p_type" in inputs:
        # 满足 <feasible_domain> 格式要求
        p_type = inputs.pop("p_type")
        kwargs = inputs
        choice_values = None
        if p_type == "categorical":
            # optuna 目前的类别元素仅支持 None, bool, int, float 和 str 类型
            #   对于其他类型的元素，比如 list 和 dict，需要替换成对应 index_ls 或者 key_ls
            #   然后再根据建议的 index 或者 key 到 list 和 dict 中取值
            assert "choices" in kwargs
            if isinstance(kwargs["choices"], (list, tuple,)) and \
                    not all([isinstance(i, (bool, int, float, str,)) or i is None for i in kwargs["choices"]]):
                choice_values = kwargs["choices"]
                kwargs["choices"] = list(range(len(kwargs["choices"])))
            elif isinstance(kwargs["choices"], (dict,)):
                choice_values = kwargs["choices"]
                kwargs["choices"] = list(kwargs["choices"].keys())

        inputs = eval(f'trial.suggest_{p_type}(name=name, **kwargs)',
                      {"trial": trial, "name": pre_name, "kwargs": kwargs})
        if choice_values is not None:
            inputs = choice_values[inputs]

    # 递归
    if isinstance(inputs, (dict,)):
        for k, v in inputs.items():
            inputs[k] = sample_from_feasible_domain(trial=trial, inputs=v, pre_name=":".join([pre_name, k]))
    elif isinstance(inputs, (list, tuple,)):
        for idx, i in enumerate(inputs):
            inputs[idx] = sample_from_feasible_domain(trial=trial, inputs=i, pre_name="@".join([pre_name, str(idx)]))

    return inputs



import numpy as np


def generate_z_pattern_indices_ls(**kwargs):
    """
        生成 从原点出发进行之字形（Z形）遍历 的下标列表
            本质上就是从原点出发，按照汉明距离（各axis的坐标之和）在 shape 对应的长方体内进行宽度优先遍历
            生成的坐标列表是 indices_ls 格式，index_ls 的具体定义参见 coordinates.convert()

        原理：
            约定：
                - 使用记号 (n,k) 表示仅考虑 shape 里面前n个axis（亦即仅对前n个axis的取值进行排列组合）时，汉明距离为k的下标列表
                - 使用记号 x | (n,k) 表示在上面所述的下标列表中的所有下标的前面加上值 x
            则有递推式：
                (n+1,k) := 0|(n,k) + 1|(n,k-1) + 2|(n,k-2) + ...
            我们以 shape=[3, 3, 2, ...] 为例演示这一过程：
                首先对于 axis=0
                    (0,0)   (0,1)   (0,2)   (0,3)
                    0       1       2       None  # 超出axis=0的取值范围
                然后根据 axis=0 的结果递推 axis=1 的情况：
                    (1,0) := 0|(0,0)
                    00
                            (1,1) := 0|(0,1) + 1|(0,0)
                            01  # 0|(0,1)
                            10  # 1|(0,0)
                                    (1,2) := 0|(0,2) + 1|(0,1) + 2|(0,0)
                                    02
                                    11
                                    20
                                           (1,3) := 0|(0,3) + 1|(0,2) + 2|(0,1) + 3|(0,0)
                                            # 0|(0,3) 不存在，忽略
                                            12
                                            21
                                            # 3|(0,0) 3超出取值范围，后面抛弃
                                                    (1,4) := 0|(0,4) + 1|(0,3) + 2|(0,2) + 3|(0,1) + 4|(0,0)
                                                    # 0|(0,4) 不存在，忽略
                                                    # 1|(0,3) 不存在，忽略
                                                    22
                                                    # 3|(0,1) 3超出取值范围，后面抛弃
                依次类推：
                    (2,0)   (2,1)   (2,2)   (2,3)   (2,4)   (2,5)
                    000     001     002     012     022     122
                            010     011     021     112
                            100     020     102     121
                                    101     111
                                    110     120

        参数：
            shape：              <list/tuple of integers>

        返回：
            indices_ls：         <nparray of nparray> 坐标列表。
    """
    # 默认参数
    paras = {
        # 必要参数
        "shape": None,
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    assert isinstance(paras["shape"], (list, tuple,)) and len(paras["shape"]) > 0

    dim_size_ls = [(dim, size) for dim, size in enumerate(paras["shape"])]
    dim_size_ls.sort(key=lambda x: x[-1])
    res = [None for _ in range(sum(paras["shape"]))]
    #
    for dim, size in dim_size_ls:
        if res[0] is None:  # init
            for i in range(size):
                res[i] = np.zeros(shape=[1, len(paras["shape"])])
                res[i][0, dim] = i
        else:  # 迭代
            for np1_k in reversed(range(sum(paras["shape"]))):
                indices = []
                # 组合公式： (n+1,k) := 0|(n,k) + 1|(n,k-1) + 2|(n,k-2) + ...
                for head, n_kp1 in enumerate(reversed(range(np1_k + 1))):
                    if head >= size:
                        break
                    if res[n_kp1] is None:
                        continue

                    res[n_kp1][:, dim] = head
                    indices.append(res[n_kp1])
                if len(indices) == 0:
                    continue
                indices = np.concatenate(indices, axis=0)
                res[np1_k] = indices
    # 整合
    res = [i for i in res if i is not None]
    res = np.concatenate(res, axis=0)
    assert len(res) == np.prod(paras["shape"])

    return res


# if __name__ == '__main__':
#     shape = [3, 3]
#     indices_ls = generate_z_pattern_indices_ls(shape=shape)
#     print(indices_ls)



import numpy as np


def get_greatest_common_divisor(n, m):
    """
        找出正整数 n 和 m 之间的最大公约数
    """
    for i in [n, m]:
        assert isinstance(i, (int, np.integer,)) and i >= 1

    # 辗转相除法求最大公约数
    while True:
        temp = m % n
        if temp == 0:  # n 为公约数
            break
        n, m = temp, n

    return n



import numpy as np


def get_ticks_of_boxes(boxes):
    """
        获取 boxes 中涉及到的坐标刻度 ticks
    """
    ticks = []
    for i in range(boxes.shape[-1]):
        tk_ls = np.asarray(sorted(list(set(boxes[..., i].flatten()))))
        ticks.append(tk_ls)

    return ticks



from .detect_collision_inside_boxes import detect_collision_inside_boxes
from .detect_collision_among_boxes_ls import detect_collision_among_boxes_ls


def detect_collision(**kwargs):
    """
        碰撞检测
            Adapt the entry of different detection functions
            当输入参数中有 boxes 时，调用 detect_collision_inside_boxes() 进行碰撞检测，此时将 boxes 中的每个 box 视为一个 item
            当输入参数中有 boxes_ls 时，调用 detect_collision_among_boxes_ls()，此时将 boxes_ls 中的每个 boxes 视为一个 item
    """
    if "boxes" in kwargs:
        return detect_collision_inside_boxes(**kwargs)
    elif "boxes_ls" in kwargs:
        return detect_collision_among_boxes_ls(**kwargs)
    else:
        raise ValueError(f"parameter boxes or boxes_ls is required!")



import numpy as np


def get_pareto_points_idx(points, directions=None):
    """
        获取位于帕累托边缘上的点的序号
            这些边缘点具有的性质：与除它之外的任意一个点相比，其总有在某一个或者多个维度上具有“优势”

        参数：
            points:             <list/tuple/np.array> 点集
                                    shape: [nums, dims]
            directions:         <list/tuple of str> 各个维度的优化方向
                                    shape: [dims]
                                    目前支持以下三种优化方向：
                                        "maximize":     越大越好
                                        "minimize":     越小越好
                                        "not_care":     不考虑该维度
                                    默认为 None, 等效于 ["maximize"]*len(points)，亦即全部维度数值越大越好
        返回：
            idx_ls:             <list of index> 位于边缘的点的序号列表
    """
    points = np.asarray(points)
    assert points.ndim == 2 and len(points) > 0
    assert directions is None or isinstance(directions, (list, tuple,)) and len(directions) == points.shape[-1]

    # 计算排序的权重
    if directions is not None:
        weights = []
        for i, direction in enumerate(directions):
            if direction == "maximize":
                weights.append(-points[:, i:i + 1])
            elif direction == "minimize":
                weights.append(points[:, i:i + 1])
            elif direction == "not_care":
                pass
        if len(weights) == 0:  # 全部都是 not_care，那就别找了
            return []
        weights = np.concatenate(weights, axis=1)
    else:
        weights = -points

    # 按权重进行排序
    order_names = tuple(f'{i}' for i in range(weights.shape[-1]))
    idx_ls = np.argsort(
        np.asarray([tuple(i) for i in weights.tolist()], dtype=[(n, weights.dtype) for n in order_names]),
        order=order_names
    )

    # 收集位于 pareto 边缘的点
    #   按上面得到的顺序从前往后过一遍 points
    #       - 累积所有维度上出现过的最大值
    #       - 若当前 point 在所有维度上比累积值要小，则抛弃该 point
    res = [idx_ls[0]]
    best = weights[idx_ls[0]][:]
    for idx in idx_ls[1:]:
        if np.any(best - weights[idx] > 0):
            best = np.min([best, weights[idx]], axis=0)
            res.append(idx)

    return res


if __name__ == '__main__':
    res = get_pareto_points_idx(points=[[1.5, 2], [2, 1], [1, 3], [3, 1]], directions=["maximize", "not_care"])
    print(res)



from .cal_cfm_iteratively_by_chunk import cal_cfm_iteratively_by_chunk
from .merge_cfm_ls import merge_cfm_ls
from .convert import convert_to_numpy


class Accumulator_for_Confusion_Matrices:
    """
        支持多种方式计算 confusion_matrices，并进行累积。

        用法：
            1. __init__( decimals_of_scores=xxx )
                在初始化阶段设置 scores/thresholds 的统计精度
            2. __call__( scores=xx, labels=xx, chunk_step=xx )
                计算 confusion_matrices
                可以多次调用，结果将进行累积
            3. get( to_numpy=xxx )
                获取结果
            4. reset( decimals_of_scores=xxx )
                重置
    """

    def __init__(self, **kwargs):
        """
            设定关键参数
            参数：
                decimals_of_scores：     integer，scores 要保留的小数位数（直接截取而非四舍五入）
                                    默认为 None，表示不对 scores 进行操作。
                                    实际上，scores的取值有非常多，
                                    为了避免 thresholds 过多，可以通过限制 scores 的小数位，来对结果进行进一步合并和去重，
                                    从而减少内存占用。
        """
        self.paras = None
        self.cfm = None
        self.reset(**kwargs)

    def reset(self, **kwargs):
        """
            重置
        """
        # 默认参数
        paras = {
            "decimals_of_scores": None,
        }
        # 获取参数
        paras.update(kwargs)
        # 校验参数
        if paras["decimals_of_scores"] is not None:
            assert isinstance(paras["decimals_of_scores"], (int,)) and paras["decimals_of_scores"] >= 0

        self.cfm = None
        self.paras = paras

    def __call__(self, scores, labels, **kwargs):
        """
            计算 confusion_matrices
            参数（详细介绍参见 cal_cfm_iteratively_by_chunk()）：
                scores:             表示模型将样本预测为正样本的置信度
                labels:             描述样本的真实标签，其中1表示正样本，0表示负样本。
                    注意： scores 和 labels 支持变量的形式，和使用 Unified_Reader_Base 包裹的形式
                chunk_step:         integer，每个分块的大小
                                        默认为 None，表示将所有输入视为一个分块，此时该函数退化为 cal_cfm()
        """
        cfm = cal_cfm_iteratively_by_chunk(scores, labels,
                                           chunk_step=kwargs.get("chunk_step", None),
                                           decimals_of_scores=self.paras["decimals_of_scores"],
                                           to_numpy=False)
        if self.cfm is not None:
            self.cfm = merge_cfm_ls([self.cfm, cfm], to_numpy=False)
        else:
            self.cfm = cfm
        return self

    def get(self, to_numpy=True):
        """
            获取结果
        """
        cfm = convert_to_numpy(self.cfm, self.paras["decimals_of_scores"]) if to_numpy else self.cfm
        return cfm



metadata = {'sep': '\t', 'title': 'this is the title', 'column_name': ['epoch', 'loss', 'model_name'],
            'column_type': ['int', 'float', 'str'], 'column_num': 3}
content = {'epoch': [0, 1, 10, 11],
           'loss': [2.31, 2.22, 12.31, 12.22],
           'model_name': ['m0.pkt', 'm1.pkt', 'm01.pkt', 'm11.pkt']}



class Cache:
    def __init__(self):
        self.primes = None
        self.primes_square = None
        self.reset()

    def reset(self):
        self.primes = [1]
        self.primes_square = [i ** 2 for i in self.primes]


cache = Cache()



#
from .factory import Face_Verification_DataSet_Factory as Factory, SUPPORT_TO_GENERATE
#
from .get_executor_ls import by_block
from .get_executor_ls import by_samples
#
from .build_generator import build_generator
from .build_iterator import build_iterator



import numpy as np
from kevin_toolbox.patches.for_numpy import linalg


def cos_similar(v0, v1, axis=-1, keepdims=False, need_normalize=True):
    if need_normalize:
        v0, v1 = linalg.normalize(v0, axis=axis, ord=2), linalg.normalize(v1, axis=axis, ord=2)
    return np.sum(v0 * v1, axis=axis, keepdims=keepdims)


if __name__ == '__main__':
    print(cos_similar(v0=np.asarray([0, 0.1]), v1=np.asarray([0.3, 0.1])))



def integrate(converters):
    """
        将 converters 整合为一个 converter
    """

    def wrap(x):
        if callable(converters):
            x = converters(x)
        elif isinstance(converters, (list, tuple,)):
            for converter in converters:
                assert callable(converter)
                x = converter(x)
        return x

    return wrap



from kevin_toolbox.developing.temperate.my_iterator import My_Iterator

if __name__ == '__main__':
    a = My_Iterator([1, 2, 3, 4, 5])
    a.set_range(1, 3)
    print(len(a))
    for i in a:
        print(i)
    print(a[1])
    print(a[2])



from setuptools import setup, find_packages

setup(
    packages=find_packages(),

    # # 默认只添加 .py 文件，如果需要添加其他文件则需要令 include_package_data 为 True
    # include_package_data=True,
    # # 需要添加的额外文件列表，比如这里就表示将添加 images/raw 下的所有 .png 文件和 images/2k 下的所有 .jpg 文件
    # package_data={
    #     "images": ['raw/*.png', '2k/*.jpg'],
    # },
    # # 需要排除的文件
    # exclude_package_data={
    #     "images": ['raw/233.png'],
    # },
    # 或者参考 https://towardsdatascience.com/the-complete-guide-for-creating-a-good-pypi-package-acb5420a03f8
    # 使用 MANIFEST.in 指定需要额外添加的文件
)



"""
    实现了部分 pytorch1.13 中的 Sequential 的功能，
    用于兼容低版本
    1.8.0及其以下：没有 self.get_submodule(name)
"""
from collections import OrderedDict
from itertools import islice
import operator

import torch
import torch.nn as nn

from kevin_toolbox.env_info import version


class My_Sequential(nn.Module):
    r"""A sequential container.
    Modules will be added to it in the order they are passed in the constructor.
    Alternatively, an ordered dict of modules can also be passed in.

    To make it easier to understand, here is a small example::

        # Example of using Sequential
        model = nn.Sequential(
                  nn.Conv2d(1,20,5),
                  nn.ReLU(),
                  nn.Conv2d(20,64,5),
                  nn.ReLU()
                )

        # Example of using Sequential with OrderedDict
        model = nn.Sequential(OrderedDict([
                  ('conv1', nn.Conv2d(1,20,5)),
                  ('relu1', nn.ReLU()),
                  ('conv2', nn.Conv2d(20,64,5)),
                  ('relu2', nn.ReLU())
                ]))
    """

    def __init__(self, *args):
        super().__init__()
        if len(args) == 1 and isinstance(args[0], OrderedDict):
            for key, module in args[0].items():
                self.add_module(key, module)
        else:
            for idx, module in enumerate(args):
                self.add_module(str(idx), module)

    def _get_item_by_idx(self, iterator, idx):
        """Get the idx-th item of the iterator"""
        size = len(self)
        idx = operator.index(idx)
        if not -size <= idx < size:
            raise IndexError('index {} is out of range'.format(idx))
        idx %= size
        return next(islice(iterator, idx, None))

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return self.__class__(OrderedDict(list(self._modules.items())[idx]))
        else:
            return self._get_item_by_idx(self._modules.values(), idx)

    def __setitem__(self, idx, module):
        key = self._get_item_by_idx(self._modules.keys(), idx)
        return setattr(self, key, module)

    def __delitem__(self, idx):
        if isinstance(idx, slice):
            for key in list(self._modules.keys())[idx]:
                delattr(self, key)
        else:
            key = self._get_item_by_idx(self._modules.keys(), idx)
            delattr(self, key)

    def __len__(self):
        return len(self._modules)

    def __dir__(self):
        keys = super(Sequential, self).__dir__()
        keys = [key for key in keys if not key.isdigit()]
        return keys

    def forward(self, input):
        for module in self._modules.values():
            input = module(input)
        return input

    def get_submodule(self, target: str) -> "Module":
        """
        Returns the submodule given by ``target`` if it exists,
        otherwise throws an error.

        For example, let's say you have an ``nn.Module`` ``A`` that
        looks like this:

        .. code-block:: text

            A(
                (net_b): Module(
                    (net_c): Module(
                        (conv): Conv2d(16, 33, kernel_size=(3, 3), stride=(2, 2))
                    )
                    (linear): Linear(in_features=100, out_features=200, bias=True)
                )
            )

        (The diagram shows an ``nn.Module`` ``A``. ``A`` has a nested
        submodule ``net_b``, which itself has two submodules ``net_c``
        and ``linear``. ``net_c`` then has a submodule ``conv``.)

        To check whether or not we have the ``linear`` submodule, we
        would call ``get_submodule("net_b.linear")``. To check whether
        we have the ``conv`` submodule, we would call
        ``get_submodule("net_b.net_c.conv")``.

        The runtime of ``get_submodule`` is bounded by the degree
        of module nesting in ``target``. A query against
        ``named_modules`` achieves the same result, but it is O(N) in
        the number of transitive modules. So, for a simple check to see
        if some submodule exists, ``get_submodule`` should always be
        used.

        Args:
            target: The fully-qualified string name of the submodule
                to look for. (See above example for how to specify a
                fully-qualified string.)

        Returns:
            torch.nn.Module: The submodule referenced by ``target``

        Raises:
            AttributeError: If the target string references an invalid
                path or resolves to something that is not an
                ``nn.Module``
        """
        if target == "":
            return self

        atoms: List[str] = target.split(".")
        mod: torch.nn.Module = self

        for item in atoms:

            if not hasattr(mod, item):
                raise AttributeError(mod._get_name() + " has no "
                                                       "attribute `" + item + "`")

            mod = getattr(mod, item)

            if not isinstance(mod, torch.nn.Module):
                raise AttributeError("`" + item + "` is not "
                                                  "an nn.Module")

        return mod


if version.compare(torch.__version__, "<=", "1.8", mode="short"):
    Sequential = My_Sequential
else:
    Sequential = nn.Sequential



from kevin_toolbox.computer_science.data_structure import Executor
from kevin_toolbox.data_flow.core.reader import Unified_Reader_Base
from kevin_toolbox.machine_learning.dataset.face.verification import Factory, SUPPORT_TO_GENERATE


def get_executor_ls_by_samples(factory, samples, **kwargs):
    """
        通过调用 verification.Factory 中的 generate_by_samples() 函数，
            来根据 samples 生成一系列的执行器 executor_ls，
            每个执行器在被 executor() 调用后都将返回一个数据集

        参数：
            samples:                list of feature_id pairs
                                        np.array with dtype=np.int
                                        shape [sample_nums, 2]
                                        需要被 Unified_Reader_Base 包裹
            factory:                verification.Factory 实例

        设定数据集大小：
            chunk_step:             每次返回的数据集的大小的上界（根据实际内存容量来选择）
            upper_bound_of_dataset_size：    chunk_step 参数的别名
                                        同时设置时，以两者最小值为准。

        输入到 Factory.generate_by_samples 中：
            need_to_generate:       需要生成的字段
                                        （参见 Factory.generate_by_samples() 中的介绍）
            feature_ids_is_sequential:       boolean，feature_ids 是否以1为间距递增的
                                        （参见 Factory.generate_by_samples() 中的介绍）

        返回：
            executor_ls：            list of get_executor_ls that can generate dataset by using get_executor_ls()
            size_ls：                产生的数据集的预期大小
    """
    "paras"
    paras = {
        "chunk_step": None,
        "upper_bound_of_dataset_size": None,
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    # samples
    assert isinstance(samples, (Unified_Reader_Base,)), \
        Exception(f"The type of input factory should be {Unified_Reader_Base}, but get a {type(samples)}!")
    # factory
    assert isinstance(factory, (Factory,)), \
        Exception(f"The type of input factory should be {Factory}, but get a {type(factory)}!")
    # chunk_step / upper_bound_of_dataset_size
    temp_ls = []
    if paras["upper_bound_of_dataset_size"] is not None:
        temp_ls.append(paras["upper_bound_of_dataset_size"])
    if paras["chunk_step"] is not None:
        temp_ls.append(paras["chunk_step"])
    assert len(temp_ls) > 0
    chunk_step = min(temp_ls)

    "body"
    executor_ls, size_ls = [], []
    # 填充参数（静态）
    kwargs__ = dict()
    for key in {"need_to_generate", "feature_ids_is_sequential"}:
        if key in kwargs:
            kwargs__[key] = kwargs[key]
    #
    count = 0
    while count < len(samples):
        # step
        step = min(chunk_step, len(samples) - count)
        # 填充参数（动态生成）
        f_kwargs = dict(samples=Executor(func=samples.read, args=[count, count + step]))
        for key in SUPPORT_TO_GENERATE:
            if key in kwargs and kwargs[key] is not None:
                f_kwargs[key] = Executor(func=kwargs[key].read, args=[count, count + step])
        # 计算
        executor_ls.append(Executor(func=factory.generate_by_samples,
                                    f_kwargs=f_kwargs, kwargs=kwargs__))
        size_ls.append(step)
        #
        count += step

    # 综合而言，
    #     除最后一个dataset以外，都有
    #     dataset_size = chunk_step
    #     最后一个dataset可能是残缺的，有
    #     0 < dataset_size <= dataset_size
    return executor_ls, size_ls



import fcntl
import os
import time
from kevin_toolbox.patches import for_os


class Mutex_Lock:
    """
        互斥锁
            基于非阻塞锁实现，可以避免死锁
    """

    def __init__(self, **kwargs):
        """
            参数：
                lock_name:              <str> string or file_path
                                            当为单纯的字符串时，将在当前文件夹下创建锁文件
                wait_interval:          <int/float> 处于阻塞状态时，获取锁的每次尝试之间的时间间隔，单位为 second。
                                            默认为 1e-1

        """
        # 默认参数
        paras = {
            # 必要参数
            "lock_name": ".233",
            "wait_interval": 1e-1,
        }

        # 获取参数
        paras.update(kwargs)

        # 校验参数
        self.paras = paras

        if os.path.dirname(self.paras["lock_name"]):
            os.makedirs(os.path.dirname(self.paras["lock_name"]), exist_ok=True)
        self.fd = open(self.paras["lock_name"], 'a')
        self.state = dict(acquired=False)

    def acquire(self, b_block_if_fail=True, time_out=None):
        start_time = time.time()
        while True:
            try:
                # 尝试获取文件锁
                fcntl.flock(self.fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
                self.state["acquired"] = True
                break
            except IOError:
                # 获取锁失败
                if not b_block_if_fail or (time_out is not None and time.time() - start_time > time_out):
                    # 超时或者设置了不阻塞
                    self.state["acquired"] = False
                    break
                else:
                    # 未超时，等待一段时间后重试
                    time.sleep(self.paras["wait_interval"])
        return self.state["acquired"]

    def release(self, b_cool_down=True):
        assert self.state["acquired"], \
            f'Please use acquire() to acquire the lock before calling release()'
        fcntl.flock(self.fd, fcntl.LOCK_UN)
        if b_cool_down:
            time.sleep(self.paras["wait_interval"] * 2)  # 冷却当前锁，让其他排队中的锁有机会获取资源，避免当前锁对资源的长期占用

    @property
    def acquired(self):
        return self.state["acquired"]

    def __del__(self):
        try:
            self.fd.close()
        except:
            pass



from abc import ABC, abstractmethod
import os
import numpy as np
from kevin_toolbox.data_flow.core.reader import File_Iterative_Reader
from kevin_toolbox.data_flow.core.cache import Cache_Manager_for_Iterator, Strategies


class Unified_Reader_Base(ABC):
    """
        按行读取数据的抽象基类
            支持通过对以
            - 内存变量
            - 持久化磁盘文件
            等多种方式保存的数据进行
            - 顺序读取
            - 随机读取

        几种读取方式之间，具有以下的优先级：
            1.内存变量
                直接用切片的方式从内存中读取。
                    相关参数：
                        var：      np.array with dtype=object
                优点：速度最快。
                缺点：内存占用大。（当数据过大无法一次性读取到内存中时，将使用下一种方法读取）
            2.持久化磁盘文件（二进制缓存）
                通过缓存管理器 Cache_Manager_for_Iterator 来进行读取
                    相关参数：
                        folder_path:    保存data的二进制缓存文件目录
                        chunk_size:     data块的大小
            3.持久化磁盘文件（数据文件）
                首先构建分批次读取文件内容的迭代器（默认使用 File_Iterative_Reader）
                然后通过适用于迭代器/生成器的缓存管理器 Cache_Manager_for_Iterator 来提高读取效率并支持随机读取。
                    相关参数：
                        file_path:      保存data的文件
                        folder_path:    保存data的二进制缓存文件目录（目录内容必须为空，否则报错）
                        chunk_size:     data块的大小
        需要实现的抽象方法：
            # 读取数据
            __get_file_iterative_reader( file_path, chunk_size )
                如何迭代读取文件到内存
                    要求返回一个迭代读取文件内容的生成器、迭代器（默认使用 File_Iterative_Reader）
            __get_cache_manager( iterator )
                （默认使用 Cache_Manager_for_Iterator）
            # 数据处理
            __deal_data( data )
                对输出前的数据进行什么处理

        数据流：
         __get_file_iterative_reader() ==> file_iterative_reader
                                                    ||
                     __get_cache_manager() ==> cache_manager
                                                    ||
                                             file =====>  file_data     var
                                                            ||          ||
                                                            ==============
                                                                    ||
                                                       __deal_data( data )
                                                                    ||
                                                                   outputs
    """

    def __init__(self, *args, **kwargs):
        """
            设定关键参数
            参数：
                var：      np.array with dtype=object
                folder_path:    保存data的二进制缓存文件目录
                file_path:      保存data的文件路径
                chunk_size:     data块的大小
                                    根据内存限制来自定义设定，默认为1k
            其他参数：

        """

        # 默认参数
        paras = {
            # 方式1
            "var": None,
            # 方式2、3
            "folder_path": None,
            "file_path": None,
            "chunk_size": 1000,
        }

        # 获取参数
        paras.update(kwargs)

        # 判断采用哪种方式
        mode = [key for key, value in {0: "var", 1: "folder_path", 2: "file_path"}.items() if
                paras.get(value, None) is not None]
        assert len(mode) > 0
        self.mode = min(mode)  # 优先选取第一种模式

        if self.mode != 0:
            # 校验参数
            paras["chunk_size"] = int(paras["chunk_size"])
            # 提前准备
            if paras["file_path"] is not None:
                assert os.path.isfile(paras["file_path"]), \
                    Exception(f"Error: file {paras['file_path']} not exists!")
                reader = self.get_file_iterative_reader(**paras)
            else:
                reader = None
            paras["manager"] = self.get_cache_manager(reader, paras["folder_path"])

        self.paras = paras

        # 其他变量
        self.index = -1

    # ------------------------------------ 读取数据 ------------------------------------ #

    @abstractmethod
    def get_file_iterative_reader(self, file_path, chunk_size, **kwargs):
        """
            如何迭代读取文件到内存
                要求返回一个按照 chunk_size 来迭代读取文件内容的生成器、迭代器（默认使用 File_Iterative_Reader）
        """
        reader = File_Iterative_Reader(file_path=file_path, chunk_size=chunk_size,
                                       drop=False, loop_num=1,
                                       convert_func=lambda x: np.array(x))
        return reader

    @abstractmethod
    def get_cache_manager(self, iterator, folder_path):
        """
            默认使用 Cache_Manager_for_Iterator
        """
        manager = Cache_Manager_for_Iterator(iterator=iterator, folder_path=folder_path,
                                             cache_update_strategy=lambda x: Strategies.drop_min_last_time(
                                                 cache_metadata=x,
                                                 cache_size_upper_bound=3))
        return manager

    def read(self, *args, **kwargs):
        if "begin" in kwargs or isinstance(args[0], (int,)):
            data = self.__read_continuously(*args, **kwargs)
        elif "id_ls" in kwargs or isinstance(args[0], (list,)):
            data = self.__read_discretely(*args, **kwargs)
        else:
            raise ValueError
        data = self.deal_data(data)

        return data

    def __read_continuously(self, begin, end=None, *args, **kwargs):
        """
            连续地读取某个区间 [begin, end)
        """
        end = begin + 1 if end is None else end
        assert begin < end

        if begin >= len(self):
            return np.array([])

        if self.mode == 0:
            data = self.paras["var"][begin:end]
        else:
            # 找出区间涉及的chunk
            chunk_ls = []
            chunk_id_beg = min(begin // self.paras["chunk_size"], max(len(self.paras["manager"]) - 1, 0))
            chunk_id_end = min(end // self.paras["chunk_size"] + 1, len(self.paras["manager"]))
            for chunk_id in range(chunk_id_beg, chunk_id_end):
                chunk_ls.append(self.paras["manager"].read(chunk_id))
            # 合并
            if len(chunk_ls) > 0:
                chunks = np.concatenate(chunk_ls, axis=0)
            else:
                chunks = np.array([])
            # 截取
            offset = begin - begin % self.paras["chunk_size"]
            data = chunks[begin - offset:end - offset]

        self.index = end - 1

        return data

    def __read_discretely(self, id_ls, *args, **kwargs):
        """
            离散地读取元素
        """
        assert max(id_ls) < len(self)

        if self.mode == 0:
            data = self.paras["var"][tuple(id_ls),]
        else:
            item_ls = [None] * len(id_ls)
            item_id_ls = np.array(id_ls)
            chunk_id_ls = item_id_ls // self.paras["chunk_size"]
            # 在涉及的chunk中，依次找出对应的 item
            for chunk_id in set(chunk_id_ls):
                chunk = self.paras["manager"].read(chunk_id)
                for i, (cid, iid) in enumerate(zip(chunk_id_ls, item_id_ls)):
                    if cid == chunk_id:
                        item_ls[i] = chunk[(iid % self.paras["chunk_size"],),]
            assert None not in item_ls, \
                Exception(f"Error: Unable to read the {item_ls.index(None)}th element!")
            # 合并
            data = np.concatenate(item_ls, axis=0)

        self.index = max(id_ls)

        return data

    # ------------------------------------ 查找数据 ------------------------------------ #

    def find(self, value, try_to_build_hash_table=True, beg_index=0):
        """
            查找元素第一次出现的序号。
                本次改进：
                原始版本中每次查找都是O(n)的，效率太低。
                因此加入通过hash表来支持检索的功能，从而将效率提高到O(1)。
            参数：
                value：              要检索的元素。
                beg_index:          从哪个序号（包含给定元素）开始检索。
                                        默认为0。
                try_to_build_hash_table:      尝试构建hash表，一次构建，后面可以反复利用，从而加速检索。
        """
        assert isinstance(beg_index, (int,)) and 0 <= beg_index < len(self), \
            f"beg_index should be integer between [0,{len(self)}), but get a {beg_index}!"

        # 尝试创建 hash 表
        if try_to_build_hash_table and "__hash_table" not in self.paras:
            try:
                self.index = -1
                hash_table = dict()
                for i, j in enumerate(self):
                    v = j[0]
                    if v not in hash_table:
                        hash_table[v] = [i]
                    else:
                        hash_table[v].append(i)
                self.paras["__hash_table"] = hash_table
                print("Hash table was successfully built.")
            except:
                print("Warn: Failed to build hash table!")

        if "__hash_table" in self.paras:
            index_ls = self.paras["__hash_table"].get(value, [])
            for index in index_ls:
                # 可以改成二分查找
                if beg_index <= index:
                    return index
        else:
            self.index = beg_index - 1
            for i, j in enumerate(self):
                if j == value:
                    return i
        return None

    # ------------------------------------ 数据处理 ------------------------------------ #

    @abstractmethod
    def deal_data(self, data):
        """
            对输出前的数据进行什么处理
        """
        return data

    # ------------------------------------ 其他 ------------------------------------ #

    # 迭代器，通过 next(self) 调用
    def __next__(self):
        if self.index < len(self) - 1:
            return self.read(self.index + 1)
        else:
            self.index = -1
            raise StopIteration

    # 支持 for 循环调用
    def __iter__(self):
        return self

    # 通过 len(self) 调用
    def __len__(self):
        if len(self.shape) > 0:
            return self.shape[0]
        else:
            return 0

    # self.shape
    @property
    def shape(self):
        if "shape" not in self.paras:
            if self.mode == 0:
                self.paras["shape"] = list(self.paras["var"].shape)
            else:
                if len(self.paras["manager"]) > 0:
                    last_chunk = self.paras["manager"].read(len(self.paras["manager"]) - 1)
                    shape = list(last_chunk.shape)
                    shape[0] += (len(self.paras["manager"]) - 1) * self.paras["chunk_size"]
                    self.paras["shape"] = shape
                else:
                    self.paras["shape"] = []
        return self.paras["shape"]



import os


class File_Iterative_Reader:
    """
        分批次读取文件内容的迭代器
        a get_generator to read the file piece by piece
    """

    def __init__(self, **kwargs):
        """
            设定关键参数
            必要参数：
                file_path:  文件路径
            读取模式相关参数：
                paras_for_open:     open() 函数的补充参数
                mode:       读取模式，默认为 "lines"
                                "lines"：  按行数计算批次大小
                                "bytes"：  按字节数计算
                chunk_size: 批次大小
                                默认为 1k
                                当为-1时，读取整个文件
                jump_size:  以后每读取一个批次就跳过的大小（每个chunk之间的gap）
                                默认为 0
                                当为-1时，跳过整个文件
                pre_jump_size: 在开始读取文件前跳过的大小
                                默认为 0
                                当为-1时，跳过整个文件
                loop_num:   循环次数
                                默认为 1 循环一次
                                小于1时，比如0、-1表示无限循环
            处理相关参数：
                drop:       丢弃不足 chunk_size 的部分。
                                默认为 False 不开启
                filter_:     过滤函数，返回一个boolean值。
                                对读取的内容进行过滤，只有过滤通过的内容才会添加到结果。
                                默认跳过空行
                map_func:   后处理函数（逐元素处理）。
                                对读取的 "lines" 中的每行或者 "bytes" 中的每个字符进行什么处理。
                                对于 "lines"，默认使用strip()去除首、尾的空格、换行符。
                                对于 "bytes"，默认不进行任何处理。
                convert_func: 后处理函数（对整体进行处理）。
                                对 "lines" 或者 "bytes" 中整体的内容进行什么处理。
                                默认不进行任何处理。
                处理流程： ==> filter_ ==> drop ==> map_func ==> convert_func
            返回：
                类型取决于 mode（"lines"返回的是列表，"bytes"是单个字符串） 以及 map_func 的处理结果
        """

        # 默认参数
        paras = {
            # 必要参数
            "file_path": None,
            # 读取模式相关参数
            "paras_for_open": dict(mode="r", encoding='utf-8'),
            "mode": "lines",
            "chunk_size": 1000,
            "jump_size": 0,
            "pre_jump_size": 0,
            "loop_num": 1,
            # 处理相关参数
            "drop": False,
            "filter_": lambda x: x != "\n",
            "map_func": lambda x: x.strip(),
            "convert_func": None,
        }

        # 获取参数
        paras.update(kwargs)

        # 校验参数
        #
        mode = paras["mode"]
        assert mode in ["lines", "bytes"]
        paras["chunk_size"] = int(paras["chunk_size"])
        paras["loop_num"] = int(paras["loop_num"]) - 1
        #
        file_path = paras["file_path"]
        assert isinstance(file_path, (str,)) and os.path.exists(file_path), \
            Exception(f"Error: file {file_path} not exists!")
        #
        paras_for_open = paras["paras_for_open"]
        assert isinstance(paras_for_open, (dict,))

        # 获取文件对象
        self.file = open(file_path, **paras_for_open)
        # 选择相应模式
        self.__read_func = {"lines": self.__read_lines, "bytes": self.__read_bytes}[mode]
        self.__jump_func = {"lines": self.__jump_lines, "bytes": self.__jump_bytes}[mode]

        self.now_in_beginning = True

        self.paras = paras

    def __next__(self):
        paras = self.paras
        while True:
            # 跳过开头部分
            if self.now_in_beginning:
                self.__jump_func(self.file, paras["pre_jump_size"], paras["filter_"])

            # 读取
            lines, end = self.__read_func(self.file, paras["chunk_size"],
                                          paras["filter_"], paras["map_func"], paras["convert_func"],
                                          paras["drop"])
            # 跳过
            self.__jump_func(self.file, paras["jump_size"], paras["filter_"])

            #
            if end:
                # 没有读到内容
                if not self.now_in_beginning and paras["loop_num"] != 0:
                    # 到达末尾，而且还有循环次数
                    # 就返回文档首部，尝试再次读取
                    self.file.seek(0)
                    self.now_in_beginning = True
                    paras["loop_num"] -= 1
                    continue
                else:
                    # 其他情况，比如：
                    #     在开头就读不到内容，是空文件
                    #     到达末尾却没有循环次数
                    # 则终止迭代
                    self.file.close()
                    raise StopIteration
            else:
                # 读到内容
                self.now_in_beginning = False
                break
        return lines

    @staticmethod
    def __read_lines(file, chunk_size, filter_, map_func, convert_func, drop):
        """
            按行读取文件中的 chunk_size 行
            返回：
                lines:      list，读取的内容
                end:        boolean，是否到文件末尾
        """
        lines = []
        while chunk_size < 0 or len(lines) < chunk_size:
            line = file.readline()
            # 到达文档最后
            if not line:
                break
            # 过滤失败
            if filter_ is not None and not filter_(line):
                continue
            # 添加
            lines.append(line)
        end = len(lines) == 0
        if not end:
            # 是否丢弃
            lines = [] if drop and len(lines) < chunk_size else lines
            # 后处理
            lines = list(map(map_func, lines)) if map_func is not None else lines
            lines = convert_func(lines) if convert_func is not None else lines
        return lines, end

    @staticmethod
    def __jump_lines(file, jump_size, filter_):
        """
            跳过 jump_size 行
            返回：
                count:      int，跳过的行数
                end:        boolean，是否到文件末尾
        """
        count = 0
        while jump_size < 0 or count < jump_size:
            line = file.readline()
            # 到达文档最后
            if not line:
                break
            # 过滤失败
            if filter_ is not None and not filter_(line):
                continue
            # 计数
            count += 1
        return count, count == 0

    @staticmethod
    def __read_bytes(file, chunk_size, filter_, map_func, convert_func, drop):
        """
            按字符数读取文件内容
            返回：
                bytes_:      string
                end:        boolean，是否到文件末尾
        """

        bytes_ = file.read(chunk_size)
        end = len(bytes_) == 0
        if not end:
            # 过滤
            bytes_ = '' if filter_ is not None and not filter_(bytes_) else bytes_
            # 是否丢弃
            bytes_ = '' if drop and len(bytes_) < chunk_size else bytes_
            # 后处理
            bytes_ = "".join([map_func(i) for i in bytes_]) if map_func is not None else bytes_
            bytes_ = convert_func(bytes_) if convert_func is not None else bytes_
        return bytes_, end

    @staticmethod
    def __jump_bytes(file, jump_size, filter_):
        """
            跳过 jump_size 个字符
            返回：
                count:      int，跳过的字符数
                end:        boolean，是否到文件末尾
        """
        bytes_ = file.read(jump_size)
        count = len(bytes_)
        return count, count == 0

    def __iter__(self):
        return self

    def __del__(self):
        try:
            del self.paras
            self.file.close()
        except Exception as e:
            print(e)


if __name__ == "__main__":
    import numpy as np

    reader = File_Iterative_Reader(file_path="developing/test_data.txt", chunk_size=2, drop=True, loop_num=2,
                                   pre_jump_size=3, convert_func=lambda x: np.array(x))
    for i in reader:
        print(i)

    del reader



import numpy as np


def get_inverse_of_transpose_index_ls(index_ls):
    """
        获取转置的逆
            对于任意 index_ls，有：
                x.transpose(*index_ls).transpose(*reverse_index_ls(index_ls)) == x
            恒成立。

        参数：
            index_ls:   <np.array> 格式具体参考 dimension.coordinates
    """
    r_index_ls = np.empty_like(index_ls)
    r_index_ls[index_ls] = np.arange(len(index_ls))
    return r_index_ls


if __name__ == '__main__':
    import numpy as np

    print(get_inverse_of_transpose_index_ls(index_ls=np.array([0, 4, 1, 5, 2, 3])))



import os
import subprocess
import argparse
import time

"""
检查包
- 若超过指定的有效期，则卸载。
"""

out_parser = argparse.ArgumentParser(description='check_validity_and_uninstall')
out_parser.add_argument('--package_name', type=str, required=True)
out_parser.add_argument('--expiration_timestamp', type=int, required=False, default=1e10)
out_parser.add_argument('--verbose', type=int, required=False, default=1)
args = out_parser.parse_args().__dict__

cur_timestamp = time.time()

if cur_timestamp > args["expiration_timestamp"]:
    ex = subprocess.Popen(f'pip uninstall {args["package_name"]} --yes', shell=True, stdout=subprocess.PIPE,
                          stderr=subprocess.STDOUT)
    out, _ = ex.communicate()
    res = out.decode().strip()
else:
    res = "still within the validity period"

if args["verbose"]:
    # print(args)
    print(res)



from abc import ABC, abstractmethod


class Iterator_Base(ABC):

    @abstractmethod
    def __iter__(self):
        return self

    @abstractmethod
    def __next__(self):
        value = 1
        return value



"""
iterable、iterator、sequence、generator之间的差别
（整理自：https://zhuanlan.zhihu.com/p/403830796）
Iterable（可迭代对象）
    实现 __iter()__ 或 __getitem()__ 方法
Sequence（序列）/Map（映射）
    实现 __getitem()__ 和 __len__() 方法
Iterator（迭代器）
    实现 __iter__() 和 __next__() 方法
get_generator（生成器）
    含有yield的函数，或者使用生成器推导式生成的生成器
"""
from .my_iterator_base import My_Iterator_Base
from .iterator_base import Iterator_Base
from .sequence_map_base import Sequence_Map_Base
from .my_iterator import My_Iterator



import torch


def my_around(x, decimals, inplace=True, floor=False):
    """
        保留到指定的小数位数。（类似于 np.around() 函数）
        round/floor to the given number of decimals.

        参数：
            x:                  input tensor
            decimals:           要保留的小数位数
        可选参数：
            floor：              采用四舍五入还是直接截取
                                    True：   直接截取
                                    False：  四舍五入
                                    默认为 False
            inplace：            boolean，是否直接在原 tensor 上进行操作
                                    默认为 True
                                    对于 requires_grad=True 的情况，此参数固定为 False
    """
    assert isinstance(decimals, (int,)) and decimals >= 0
    assert isinstance(x, torch.Tensor)

    if not inplace or x.requires_grad:
        x = x.clone()

    scale = 10 ** decimals
    x *= scale
    if floor:
        x.floor_()
    else:
        x.round_()
    x /= scale
    return x


if __name__ == '__main__':
    a = torch.Tensor([1, 2.4342, 8.1234567])
    for inplace in [False, True]:
        a_r = my_around(a, decimals=3, inplace=inplace)
        print(f"value a {a} a_r {a_r}")
        print(f"id a {id(a)} a_r {id(a_r)}")



def deep_update(stem, patch):
    """
        使用 patch 递归地更新 stem 中对应结构下的值
            比如对于 stem = {"a": {"b": [233], "g": 3}}
            如果使用 patch = {"a": {"b": 444}} 进行更新，可以得到：
                {'a': {'b': 444, 'g': 3}}
    """
    assert isinstance(stem, (dict,)) and isinstance(patch, (dict,))

    return _recursion(stem=stem, patch=patch)


def _recursion(stem, patch):
    for k, v in patch.items():
        if k in stem and isinstance(v, (dict,)):
            stem[k] = _recursion(stem=stem[k], patch=v)
        else:
            stem[k] = v
    return stem


if __name__ == '__main__':
    stem = {"a": {"b": [233], "g": 3}}
    patch = {"a": {"b": 444}}
    print(deep_update(stem, patch))



def cal_tpr_and_fpr(cfm):
    """
        根据混淆矩阵来计算出各个 threshold 下的 tpr、fpr

        计算 tpr、fpr 的原理：
            tpr = TP/TP+FN  # 正样本中被模型检出的比例，越大越好
            fpr = FP/FP+TN  # 负样本中被模型误检的比例，越小越好

        背景知识：
            假设前 m 个满足阈值，则有 TP+FP == m，且 TP、FP 都是单调递增的，
            因此，fpr 和 tpr 都是随着m增加而递增的，
            我们希望在两者之间取一个平衡点，亦即 threshold。
            一般是对于给定 fpr 时，求出此时的 tpr和 threshold，
            记为 tpr=xxx@fpr=xxx 和 threshold=xxx@fpr=xxx
    """
    cfm["tpr_ls"] = cfm["tp_ls"] / (cfm["tp_ls"] + cfm["fn_ls"] + 1e-14)
    cfm["fpr_ls"] = cfm["fp_ls"] / (cfm["fp_ls"] + cfm["tn_ls"] + 1e-14)
    return cfm



import json
from kevin_toolbox.data_flow.file.json_.converter import integrate


def read_json(file_path, converters=None):
    with open(file_path, 'r') as f:
        obj = json.load(f, object_hook=integrate(converters))
    return obj



import numpy as np


def spilt_integer_most_evenly(x, group_nums):
    assert isinstance(x, (int, np.integer,)) and x >= 0 and group_nums > 0

    res = np.ones(group_nums, dtype=int) * (x // group_nums)
    res[:x % group_nums] += 1
    return res.tolist()


if __name__ == '__main__':
    print(spilt_integer_most_evenly(x=100, group_nums=7))




def general_matrix_multiplication():
    """
        广义通用矩阵乘法操作
    """
    x = tf.cast(x, dtype=tf.float32)
    size = list(x.shape)
    pre_size, array_len, dims_of_point = size[:-2], size[-2], size[-1]

    # x_tile_lens [..., array_len * array_len, dims_of_point]
    x_tile_lens = tf.reshape(tf.tile(x, [1] * len(pre_size) + [array_len, 1]),
                             shape=pre_size + [array_len * array_len, dims_of_point])
    x_tile_dims = tf.reshape(tf.tile(x, [1] * len(pre_size) + [1, array_len]),
                             shape=pre_size + [array_len * array_len, dims_of_point])
    if return_type == "matrix":
        # distances [..., array_len * array_len, 1] ==> distance_matrix [..., array_len, array_len]
        distances = cal_distance(x_tile_dims, x_tile_lens)
        return tf.reshape(distances, shape=pre_size + [array_len, array_len])
    elif return_type == "array":
        mask = tf.linalg.band_part(
            tf.ones(shape=[array_len, array_len], dtype=tf.int8) - tf.eye(array_len, dtype=tf.int8),
            num_lower=0, num_upper=-1)  # 上三角1矩阵
        mask = tf.reshape(mask, shape=[-1])
        mask_index = tf.where(mask)
        # 将最后的两个维度提前
        index = list(range(len(x_tile_lens.shape)))
        target_index = index[-2:] + index[:-2]
        x_tile_lens = tf.transpose(x_tile_lens, perm=target_index)
        x_tile_dims = tf.transpose(x_tile_dims, perm=target_index)
        # 只保留上三角
        x_tile_lens = tf.gather_nd(params=x_tile_lens, indices=mask_index)
        x_tile_dims = tf.gather_nd(params=x_tile_dims, indices=mask_index)
        # 恢复维度顺序
        target_index = index[2:] + index[:2]
        x_tile_lens = tf.transpose(x_tile_lens, perm=target_index)
        x_tile_dims = tf.transpose(x_tile_dims, perm=target_index)
        # 计算距离
        distances = cal_distance(x_tile_dims, x_tile_lens)
        return distances[..., 0]
    elif return_type == "__test":
        print(tf.matmul(x, x, transpose_b=bool))
    else:
        raise TypeError


class Node:
    def __init__(self, **kwargs):
        """
            设定关键参数

            参数：
                file_path:  文件路径
            读取模式相关参数：
                paras_for_open:     open() 函数的补充参数
                mode:       读取模式，默认为 "lines"
                                "lines"：  按行数计算批次大小
                                "bytes"：  按字节数计算
                chunk_size: 批次大小
                                默认为 1k
                                当为-1时，读取整个文件
                jump_size:  以后每读取一个批次就跳过的大小（每个chunk之间的gap）
                                默认为 0
                                当为-1时，跳过整个文件
                pre_jump_size: 在开始读取文件前跳过的大小
                                默认为 0
                                当为-1时，跳过整个文件
                loop_num:   循环次数
                                默认为 1 循环一次
                                小于1时，比如0、-1表示无限循环
            处理相关参数：
                drop:       丢弃不足 chunk_size 的部分。
                                默认为 False 不开启
                filter_:     过滤函数，返回一个boolean值。
                                对读取的内容进行过滤，只有过滤通过的内容才会添加到结果。
                                默认跳过空行
                map_func:   后处理函数。
                                对读取的 "lines" 或者 "bytes" 进行什么处理。
                                默认使用strip()去除首、尾的空格、换行符。
                处理流程： ==> filter_ ==> drop ==> map_func ==>
            返回：
                类型取决于 mode（"lines"返回的是列表，"bytes"是单个字符串） 以及 map_func 的处理结果
        """

        # 默认参数
        paras = {
            #
            "uid": None,
        }

        # 获取参数
        paras.update(kwargs)

        # 校验参数
        assert paras["uid"] in ["lines", "bytes"]
        paras["chunk_size"] = int(paras["chunk_size"])



import torch
import numpy as np
from kevin_toolbox.env_info import version

if version.compare(torch.__version__, "<", "1.2", mode="short"):
    """
        参考 torch.where()
            在 1.1 版本及其之前的 pytorch 中的 torch.where()，
            只有 torch.where(condition, x, y) 的用法，
            而没有 torch.where(condition) 的用法，
            考虑到兼容性，在pytorch版本过低时，自动使用 numpy.where() 来替代
    """


    def where(*args):
        if len(args) == 1:
            res = np.where(args[0].cpu())
            return tuple(torch.as_tensor(i, device=args[0].device) for i in res)
        else:
            return torch.where(*args)
else:
    where = torch.where



import copy
import torch
import numpy as np
from kevin_toolbox.computer_science.algorithm.statistician._init_var import _init_var


class Exponential_Moving_Average:
    """
        滑动平均器
            支持为每个输入数据配置不同的权重
    """

    def __init__(self, **kwargs):
        """
            参数：
                keep_ratio:             <float> 对历史值的保留比例。
                                            其意义为： 大致等于计算过去 1/keep_ratio 个数据的平均值
                                            默认为 0.99
                                            当设置为 0 时相当于仅保留最新的数据
                bias_correction:        <boolean> 是否开启偏差修正。
                                            默认为 True
                update_func:            <function> 用于融合新旧数据的函数。
                                            默认为：
                                            lambda w_old, v_old, w_new, v_new: w_old * v_old + w_new * v_new
                                            你可以利用该接口，指定你需要的融合方式，比如融合后对数据进行归一化：
                                            lambda w_old, v_old, w_new, v_new: normalize(w_old * v_old + w_new * v_new)
                指定输入数据的格式，有三种方式：
                    1. 显式指定数据的形状和所在设备等。
                        data_format:        <dict of paras>
                                其中需要包含以下参数：
                                    type_:              <str>
                                                            "numpy":        np.ndarray
                                                            "torch":        torch.tensor
                                    shape:              <list of integers>
                                    device:             <torch.device>
                                    dtype:              <torch.dtype>
                    2. 根据输入的数据，来推断出形状、设备等。
                        like:               <torch.tensor / np.ndarray / int / float>
                    3. 均不指定 data_format 和 like，此时将等到第一次调用 add()/add_sequence() 时再根据输入来自动推断。
                    以上三种方式，默认选用最后一种。
                    如果三种方式同时被指定，则优先级与对应方式在上面的排名相同。
        """

        # 默认参数
        paras = {
            # 超参数
            "keep_ratio": 0.99,
            "bias_correction": True,
            # 指定累加方式
            "update_func": lambda w_old, v_old, w_new, v_new: w_old * v_old + w_new * v_new,
            # 指定输入数据的形状、设备
            "data_format": None,
            "like": None,
        }

        # 获取参数
        paras.update(kwargs)

        # 校验参数
        assert isinstance(paras["keep_ratio"], (int, float,)) and 0 <= paras["keep_ratio"] <= 1
        #
        self.paras = paras
        self.var = _init_var(like=paras["like"], data_format=paras["data_format"])
        self.state = dict(
            total_nums=0,
            bias_fix=1,
        )

    def add_sequence(self, var_ls, weight_ls=None):
        if weight_ls is not None:
            if isinstance(weight_ls, (int, float,)):
                weight_ls = [weight_ls] * len(var_ls)
            assert len(weight_ls) == len(var_ls)
            for var, weight in enumerate(var_ls, weight_ls):
                self.add(var, weight)
        else:
            for var in var_ls:
                self.add(var)

    def add(self, var, weight=1):
        """
            添加单个数据

            参数:
                var:                数据
                weight:             <int/float> 权重。
                                        默认为 1
        """
        if self.var is None:
            self.var = _init_var(like=var)
        new_ratio = (1 - self.paras["keep_ratio"]) * weight
        keep_ratio = (1 - new_ratio)
        # 累积
        self.var = self.paras["update_func"](keep_ratio, self.var, new_ratio, var)
        #
        self.state["total_nums"] += 1
        self.state["bias_fix"] *= keep_ratio

    def get(self, bias_correction=None):
        """
            获取当前累加值
                当未初始化时，返回 None

            参数:
                bias_correction:        <boolean> 是否开启偏差修正。
                                            默认使用初始化时设定的值
        """
        if self.var is None:  # 未初始化
            return None
        bias_correction = self.paras["bias_correction"] if bias_correction is None else bias_correction
        if bias_correction:
            return self.var / (1 - self.state["bias_fix"] + 1e-10)
        else:
            return self.var

    def clear(self):
        self.var = _init_var(like=self.var)
        self.state = dict(
            total_nums=0,
            bias_fix=1,
        )

    def __len__(self):
        return self.state["total_nums"]


if __name__ == '__main__':
    seq = list(torch.tensor(range(1, 10)))
    wls = np.asarray([0.1] * 5 + [0.9] + [0.1] * 4) * 0.1
    ema = Exponential_Moving_Average(keep_ratio=0.9, bias_correction=True)
    for i, (v, w) in enumerate(zip(seq, wls)):
        ema.add(var=v, weight=w)
        print(i, v, ema.get(), ema.state["bias_fix"])



import numpy as np
import torch

REGISTERED_TYPES = dict(
    float32=dict(dtype=dict(
        numpy=np.float32,
        torch=torch.float32,
    )),
    int8=dict(dtype=dict(
        numpy=np.int8,
        torch=torch.int8,
    ), range=[-128, 127]),
    uint8=dict(dtype=dict(
        numpy=np.uint8,
        torch=torch.uint8,
    ), range=[0, 255]),
)


def convert_dtype(x, target_type):
    """
        转换 dtype 数据类型
            本函数相较于 numpy 或者 pytorch 内置的转换函数，添加了根据类型自动裁剪的步骤，从而能够避免潜在的溢出情况。
            建议使用本函数替代内置的转换函数。

        参数：
            x:          <np.array/torch.tensor>
            dtype:      <string> 转换的目标类型
                            已支持的类型：
                                "float32", "int8", "uint8"
    """
    assert isinstance(target_type, (str,)) and target_type in REGISTERED_TYPES, \
        f"currently supported types are: {set(REGISTERED_TYPES.keys())}, but got a {target_type}"

    temp = REGISTERED_TYPES[target_type]
    if type(x) is np.ndarray:
        # clip
        y = x
        if "range" in temp:
            y = np.clip(y, temp["range"][0], temp["range"][1])
        # convert
        y = y.astype(dtype=temp["dtype"]["numpy"])
    elif torch.is_tensor(x):
        # clip
        y = x
        if "range" in temp:
            y = torch.clamp(y, temp["range"][0], temp["range"][1])
        # convert
        y = y.to(dtype=temp["dtype"]["torch"])
    else:
        raise TypeError(f"type of x should be np.ndarray or torch.tensor, but get a {type(x)}")

    return y


if __name__ == '__main__':
    from line_profiler import LineProfiler

    lp = LineProfiler()
    lp_wrapper = lp(convert_dtype)
    x = torch.rand([1, 3, 928, 11200], device=torch.device("cuda"))
    lp_wrapper(x=x, target_type="uint8")
    lp.print_stats()
